#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <stdio.h>
#include <unistd.h>

#define FWD_PORT 256
#define FWD_VERSION 0x31000000

static int debug_switch = 0;

void debug_lib(void)
{
  debug_switch = 1;
}

void hex_dump(char *header, char *buff, int len)
{
  int i, k;

  printf("%s\n", header);

  for (i = 0; i < len; i += 16) {
    printf("%04x: ", i);
    for (k = 0; k < 16 && i + k < len; k++)
      printf("%02x ", (unsigned char)buff[i + k]);
    while (k++ < 16)
      printf("   ");
    for (k = 0; k < 16 && i + k < len; k++)
      if (buff[i + k] >= 32 && buff[i + k] <= 126)
	printf("%c", buff[i + k]);
      else
	printf(".");
    printf("\n");
  }
}

int full_read(int s, char *buff, int len)
{
  int res;
  char *orig_buff = buff;
  int orig_len = len;

  while (len > 0)
    if ((res = read(s, buff, len)) <= 0) {
      fprintf(stderr, "Cannot read from socket\n");
      return -1;
    } else {
      buff += res;
      len -= res;
    }

  if (debug_switch)
    hex_dump("READ", orig_buff, orig_len);

  return 0;
}
    
int full_write(int s, char *buff, int len)
{
  int res;
  char *orig_buff = buff;
  int orig_len = len;

  while (len > 0)
    if ((res = write(s, buff, len)) <= 0) {
      fprintf(stderr, "Cannot write to socket\n");
      return -1;
    } else {
      buff += res;
      len -= res;
    }

  if (debug_switch)
    hex_dump("WRITE", orig_buff, orig_len);

  return 0;
}

int write_long(int s, unsigned long data)
{
  data = htonl(data);
  return full_write(s, (char *)&data, sizeof (data));
}

int write_addr(int s, unsigned long addr)
{
  return full_write(s, (char *)&addr, sizeof (addr));
}

int read_long(int s, unsigned long *data)
{
  if (full_read(s, (char *)data, sizeof (*data)) < 0)
    return -1;

  *data = ntohl(*data);
  return 0;
}

int read_addr(int s, unsigned long *addr)
{
  return full_read(s, (char *)addr, sizeof (*addr));
}

char *asc_addr(unsigned long addr)
{
  static int idx = 0;
  char *res;
  static char buff[4][16];
  unsigned char *asc = (unsigned char *)&addr;

  res = buff[idx++];
  idx &= 3;

  sprintf(res, "%d.%d.%d.%d", asc[0], asc[1], asc[2], asc[3]);
  return res;
}

static int is_hex_digit(char c)
{
  return (c >= 'a' && c <= 'f') ||
    (c >= 'A' && c <= 'F') ||
    (c >= '0' && c <= '9');
}

unsigned char *find_hex(unsigned char *msg, unsigned long mlen, int slen)
{
  int i, k;

  for (i = 0; i < mlen - slen; i++) {
    for (k = 0; k < slen && is_hex_digit(msg[i + k]); k++);
    if (k == slen)
      break;
  }
  
  if (i < mlen - slen)
    return &msg[i];

  return "--- not found ---";
}

int part_one(unsigned long fw_addr, unsigned long peer_addr,
	     unsigned long opcode, char *auth)
{
  struct protoent *prot;
  struct sockaddr_in addr;
  unsigned long serv_addr;
  unsigned char string_data[512];
  unsigned long long_data;
  int s, i;

  if ((prot = getprotobyname("tcp")) == NULL) {
    fprintf(stderr, "Cannot get protocol number for TCP\n");
    return -1;
  }

  if ((s = socket(PF_INET, SOCK_STREAM, prot->p_proto)) < 0) {
    perror("Cannot create socket");
    return -1;
  }

  memset(&addr, 0, sizeof (addr));
  addr.sin_family = PF_INET;
  addr.sin_port = htons(FWD_PORT);
  addr.sin_addr.s_addr = fw_addr;

  if (connect(s, &addr, sizeof (addr)) < 0) {
    perror("Cannot connect socket");
    close(s);
    return -1;
  }

  if (write_long(s, FWD_VERSION) < 0) {
    close(s);
    return -2;
  }

  printf("-> Version: 0x%.8x\n", FWD_VERSION);

  if (write_long(s, opcode) < 0) {
    close(s);
    return -2;
  }

  printf("-> Command: 0x%.8lx\n", opcode);

  if (write_long(s, sizeof (peer_addr)) < 0) {
    close(s);
    return -2;
  }

  if (write_addr(s, peer_addr) < 0) {
    close(s);
    return -2;
  }

  printf("-> IP address #1: %s\n", asc_addr(peer_addr));

  if (read_long(s, &long_data) < 0) {
    close(s);
    return -2;
  }

  printf("<- Version: 0x%.8lx\n", long_data);

  if (read_long(s, &long_data) < 0) {
    close(s);
    return -2;
  }

  for (i = 0; i < long_data / sizeof (serv_addr); i++) {
    if (read_addr(s, &serv_addr) < 0) {
      close(s);
      return -2;
    }

    printf("<- IP address #%d: %s\n", i + 1, asc_addr(serv_addr));
  }

  if (read_long(s, &long_data) < 0) {
    close(s);
    return -2;
  }

  if (long_data > sizeof (string_data)) {
    fprintf(stderr, "String is too long (%ld bytes)\n", long_data);
    return -2;
  }

  if (full_read(s, string_data, long_data) < 0) {
    close(s);
    return -2;
  }

  printf("<- Authentication method: '%s'\n", string_data);

  if (strcmp(auth, string_data)) {
    fprintf(stderr,
	    "Unsupported authentication method: '%s'\n", string_data);
    close(s);
    return 3;
  }

  return s;
}

int part_two(int s)
{
  unsigned long long_data;

  if (write_long(s, 1) < 0) {
    close(s);
    return -1;
  }

  if (write_long(s, 1) < 0) {
    close(s);
    return -1;
  }

  if (full_write(s, "", 1) < 0) {
    close(s);
    return -1;
  }

  printf("-> Scope information\n");

  if (write_long(s, 1) < 0) {
    close(s);
    return -1;
  }

  if (write_long(s, 1) < 0) {
    close(s);
    return -1;
  }

  if (full_write(s, "", 1) < 0) {
    close(s);
    return -1;
  }

  printf("-> Interface information\n");

  if (read_long(s, &long_data) < 0) {
    close(s);
    return -1;
  }

  printf("<- Return code: %ld\n", long_data);

  if (long_data != 0) {
    fprintf(stderr, "Unload failed\n");
    close(s);
    return -1;
  }

  if (write_long(s, 0) < 0) {
    close(s);
    return -1;
  }

  printf("-> Command 0\n");

  return 0;
}
